package net.knarcraft.knarlib.formatting;

import net.knarcraft.knarlib.property.ColorConversion;
import net.knarcraft.knarlib.util.ColorHelper;
import net.md_5.bungee.api.ChatColor;
import org.bukkit.command.CommandSender;
import org.jetbrains.annotations.NotNull;

import java.util.List;

/**
 * A formatter for formatting displayed messages
 */
@SuppressWarnings("unused")
public final class StringFormatter {

    private final String pluginName;
    private final Translator translator;
    private ChatColor successColor = ChatColor.GREEN;
    private ChatColor errorColor = ChatColor.DARK_RED;
    private String namePrefix = "[";
    private String nameSuffix = "]";
    private ColorConversion colorConversion = ColorConversion.NORMAL;

    /**
     * Instantiates a new string formatter
     *
     * <p>You may leave one of these as null if not in use, for example if you don't use translatable strings.</p>
     *
     * @param pluginName <p>The name of the using plugin (used for the message prefix)</p>
     * @param translator <p>The translator to use for translating messages</p>
     */
    public StringFormatter(@NotNull String pluginName, @NotNull Translator translator) {
        this.pluginName = pluginName;
        this.translator = translator;
    }

    /**
     * Sets the type of color conversion to use for translated messages
     *
     * @param colorConversion <p>The color conversion to use</p>
     */
    public void setColorConversion(@NotNull ColorConversion colorConversion) {
        this.colorConversion = colorConversion;
    }

    /**
     * Sets the prefix used in front of the plugin name. Normally just "["
     *
     * @param namePrefix <p>The new name prefix</p>
     */
    public void setNamePrefix(@NotNull String namePrefix) {
        this.namePrefix = namePrefix;
    }

    /**
     * Sets the suffix used after the plugin name. Normally just "]"
     *
     * @param nameSuffix <p>The new name suffix</p>
     */
    public void setNameSuffix(@NotNull String nameSuffix) {
        this.nameSuffix = nameSuffix;
    }

    /**
     * Sets the color to prepend when using displaySuccessMessage
     *
     * @param color <p>The color to use</p>
     */
    public void setSuccessColor(@NotNull ChatColor color) {
        this.successColor = color;
    }

    /**
     * Sets the color to prepend when using displayErrorMessage
     *
     * @param color <p>The color to use</p>
     */
    public void setErrorColor(@NotNull ChatColor color) {
        this.errorColor = color;
    }

    /**
     * Displays a formatted message
     *
     * @param sender  <p>The command sender to display the message to</p>
     * @param message <p>The translatable message to display</p>
     */
    public void displayNeutralMessage(@NotNull CommandSender sender, @NotNull TranslatableMessage message) {
        displayNeutralMessage(sender, this.translator.getTranslatedMessage(message));
    }

    /**
     * Displays a formatted message
     *
     * @param sender  <p>The command sender to display the message to</p>
     * @param message <p>The raw message to display</p>
     */
    public void displayNeutralMessage(@NotNull CommandSender sender, @NotNull String message) {
        sender.sendMessage(getFormattedMessage(message));
    }

    /**
     * Displays a message signifying a successful action
     *
     * @param sender  <p>The command sender to display the message to</p>
     * @param message <p>The translatable message to display</p>
     */
    public void displaySuccessMessage(@NotNull CommandSender sender, @NotNull TranslatableMessage message) {
        displaySuccessMessage(sender, this.translator.getTranslatedMessage(message));
    }

    /**
     * Displays a message signifying a successful action
     *
     * @param sender  <p>The command sender to display the message to</p>
     * @param message <p>The raw message to display</p>
     */
    public void displaySuccessMessage(@NotNull CommandSender sender, @NotNull String message) {
        if (!this.namePrefix.equals("[") || !this.nameSuffix.equals("]")) {
            sender.sendMessage(getFormattedMessage(successColor + message));
        } else {
            sender.sendMessage(successColor + getFormattedMessage(message));
        }
    }

    /**
     * Displays a message signifying an unsuccessful action
     *
     * @param sender  <p>The command sender to display the message to</p>
     * @param message <p>The translatable message to display</p>
     */
    public void displayErrorMessage(@NotNull CommandSender sender, @NotNull TranslatableMessage message) {
        displayErrorMessage(sender, this.translator.getTranslatedMessage(message));
    }

    /**
     * Displays a message signifying an unsuccessful action
     *
     * @param sender  <p>The command sender to display the message to</p>
     * @param message <p>The raw message to display</p>
     */
    public void displayErrorMessage(@NotNull CommandSender sender, @NotNull String message) {
        if (!this.namePrefix.equals("[") || !this.nameSuffix.equals("]")) {
            sender.sendMessage(getFormattedMessage(errorColor + message));
        } else {
            sender.sendMessage(errorColor + getFormattedMessage(message));
        }
    }

    /**
     * Gets the raw, unformatted translation of the given translatable message
     *
     * @param translatableMessage <p>The message to get the translation of</p>
     * @return <p>The raw translation</p>
     */
    public @NotNull String getUnformattedMessage(@NotNull TranslatableMessage translatableMessage) {
        return this.translator.getTranslatedMessage(translatableMessage);
    }

    /**
     * Gets the unformatted translation of the given translatable message, with translated color codes
     *
     * @param translatableMessage <p>The message to get the colored translation of</p>
     * @return <p>The colored raw translation</p>
     */
    public @NotNull String getUnformattedColoredMessage(@NotNull TranslatableMessage translatableMessage) {
        return ColorHelper.translateColorCodes(this.translator.getTranslatedMessage(translatableMessage),
                this.colorConversion);
    }

    /**
     * Replaces a placeholder in a translatable message
     *
     * @param translatableMessage <p>The translatable message to replace in</p>
     * @param placeholder         <p>The placeholder to replace</p>
     * @param replacement         <p>The replacement value</p>
     * @return <p>The input string with the placeholder replaced</p>
     */
    public String replacePlaceholder(@NotNull TranslatableMessage translatableMessage, @NotNull String placeholder,
                                     @NotNull String replacement) {
        return replacePlaceholder(this.translator.getTranslatedMessage(translatableMessage), placeholder, replacement);
    }

    /**
     * Replaces placeholders in a translatable message
     *
     * @param translatableMessage <p>The translatable message to replace in</p>
     * @param placeholders        <p>The placeholders to replace</p>
     * @param replacements        <p>The replacement values</p>
     * @return <p>The input string with placeholders replaced</p>
     */
    public String replacePlaceholders(@NotNull TranslatableMessage translatableMessage, @NotNull String[] placeholders,
                                      @NotNull String[] replacements) {
        return replacePlaceholders(this.translator.getTranslatedMessage(translatableMessage), List.of(placeholders),
                List.of(replacements));
    }

    /**
     * Replaces placeholders in a translatable message
     *
     * @param translatableMessage <p>The translatable message to replace in</p>
     * @param placeholders        <p>The placeholders to replace</p>
     * @param replacements        <p>The replacement values</p>
     * @return <p>The input string with placeholders replaced</p>
     */
    public String replacePlaceholders(@NotNull TranslatableMessage translatableMessage, @NotNull List<String> placeholders,
                                      @NotNull List<String> replacements) {
        return replacePlaceholders(this.translator.getTranslatedMessage(translatableMessage), placeholders, replacements);
    }

    /**
     * Replaces a placeholder in a string
     *
     * @param input       <p>The input string to replace in</p>
     * @param placeholder <p>The placeholder to replace</p>
     * @param replacement <p>The replacement value</p>
     * @return <p>The input string with the placeholder replaced</p>
     */
    public static String replacePlaceholder(@NotNull String input, @NotNull String placeholder,
                                            @NotNull String replacement) {
        return input.replace(placeholder, replacement);
    }

    /**
     * Replaces placeholders in a string
     *
     * @param input        <p>The input string to replace in</p>
     * @param placeholders <p>The placeholders to replace</p>
     * @param replacements <p>The replacement values</p>
     * @return <p>The input string with placeholders replaced</p>
     */
    public static String replacePlaceholders(@NotNull String input, @NotNull String[] placeholders,
                                             @NotNull String[] replacements) {
        return replacePlaceholders(input, List.of(placeholders), List.of(replacements));
    }

    /**
     * Replaces placeholders in a string
     *
     * @param input        <p>The input string to replace in</p>
     * @param placeholders <p>The placeholders to replace</p>
     * @param replacements <p>The replacement values</p>
     * @return <p>The input string with placeholders replaced</p>
     */
    public static String replacePlaceholders(@NotNull String input, @NotNull List<String> placeholders,
                                             @NotNull List<String> replacements) {
        for (int i = 0; i < Math.min(placeholders.size(), replacements.size()); i++) {
            input = replacePlaceholder(input, placeholders.get(i), replacements.get(i));
        }
        return input;
    }

    /**
     * Gets the formatted version of any chat message
     *
     * @param message <p>The message to format</p>
     * @return <p>The formatted message</p>
     */
    private String getFormattedMessage(@NotNull String message) {
        message = ColorHelper.translateColorCodes(message, this.colorConversion);
        String coloredPrefix = ColorHelper.translateColorCodes(namePrefix + this.pluginName +
                nameSuffix, this.colorConversion);
        return coloredPrefix + " " + ChatColor.RESET + message;
    }

}
