package net.knarcraft.knarlib.formatting;

import net.knarcraft.knarlib.property.ColorConversion;
import net.knarcraft.knarlib.util.ColorHelper;
import net.knarcraft.knarlib.util.FileHelper;
import org.bukkit.Bukkit;
import org.bukkit.configuration.file.YamlConfiguration;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

/**
 * A tool to get strings translated to the correct language
 *
 * <p>To add translatable messages, create a class extending TranslatableMessage, and register it by calling
 * "registerMessageCategory" on one of its values.
 * Add a file strings.yml in your resources directory with data: `en.ENUM: "Text"`. You must have a `en` language as
 * it's used as the fallback, but you can additionally add any language code you want.</p>
 */
@SuppressWarnings("unused")
public final class Translator {

    private final List<TranslatableMessage> messageCategories = new ArrayList<>();
    private Map<TranslatableMessage, String> translatedMessages;
    private Map<TranslatableMessage, String> backupTranslatedMessages;
    private ColorConversion colorConversion = ColorConversion.RGB;

    /**
     * Registers all translatable messages in the given message category
     *
     * <p>This should be run for one enum of every class extending TranslatableMessage. This allows the translator to
     * look for translations for any enum in the category.</p>
     *
     * @param translatableMessage <p>A translatable message in the category to register</p>
     */
    public void registerMessageCategory(@NotNull TranslatableMessage translatableMessage) {
        messageCategories.add(translatableMessage);
    }

    /**
     * Sets the color conversion to use for translated messages
     *
     * <p>See {@link net.knarcraft.knarlib.property.ColorConversion} and
     * {@link net.knarcraft.knarlib.util.ColorHelper#translateColorCodes(String, ColorConversion)} for details, but
     * basically, this allows setting which color codes should be converted, and which should not be.</p>
     *
     * @param colorConversion <p>The color conversion to be used</p>
     */
    public void setColorConversion(@NotNull ColorConversion colorConversion) {
        this.colorConversion = colorConversion;
    }

    /**
     * Loads the languages used by this translator
     *
     * @param dataFolder       <p>The data folder to load custom strings from</p>
     * @param fallbackLanguage <p>The default language to fall back on in the case of missing translations</p>
     * @param selectedLanguage <p>The currently selected language</p>
     */
    public void loadLanguages(@NotNull File dataFolder, @NotNull String fallbackLanguage,
                              @NotNull String selectedLanguage) {
        backupTranslatedMessages = loadTranslatedMessages(fallbackLanguage);
        translatedMessages = loadCustomTranslatedMessages(dataFolder, selectedLanguage);
        if (translatedMessages == null) {
            translatedMessages = loadTranslatedMessages(selectedLanguage);
        }
    }

    /**
     * Checks whether the given translatable message has a translation
     *
     * @param translatableMessage <p>The translatable message to check</p>
     * @return <p>True if the message has a translation</p>
     */
    public boolean hasTranslation(@NotNull TranslatableMessage translatableMessage) {
        return backupTranslatedMessages.containsKey(translatableMessage) &&
                !backupTranslatedMessages.get(translatableMessage).isBlank();
    }

    /**
     * Gets a translated version of the given translatable message
     *
     * @param translatableMessage <p>The message to translate</p>
     * @return <p>The translated message</p>
     */
    public @NotNull String getTranslatedMessage(@NotNull TranslatableMessage translatableMessage) {
        if (translatedMessages == null) {
            return "Translated strings not loaded";
        }
        String translatedMessage;
        if (translatedMessages.containsKey(translatableMessage)) {
            translatedMessage = translatedMessages.get(translatableMessage);
        } else if (backupTranslatedMessages.containsKey(translatableMessage)) {
            translatedMessage = backupTranslatedMessages.get(translatableMessage);
        } else {
            translatedMessage = translatableMessage.toString();
        }
        return ColorHelper.translateColorCodes(translatedMessage, colorConversion);
    }

    /**
     * Loads all translated messages for the given language
     *
     * @param language <p>The language chosen by the user</p>
     * @return <p>A mapping of all strings for the given language</p>
     */
    public @Nullable Map<TranslatableMessage, String> loadTranslatedMessages(@NotNull String language) {
        try {
            BufferedReader reader = FileHelper.getBufferedReaderForInternalFile("/strings.yml");
            return loadTranslatableMessages(language, reader);
        } catch (FileNotFoundException e) {
            Bukkit.getLogger().log(Level.SEVERE, "Unable to load translated messages");
            return null;
        }
    }

    /**
     * Tries to load translated messages from a custom strings.yml file
     *
     * @param dataFolder <p>The data folder to load custom strings from</p>
     * @param language   <p>The selected language</p>
     * @return <p>The loaded translated strings, or null if no custom language file exists</p>
     */
    public @Nullable Map<TranslatableMessage, String> loadCustomTranslatedMessages(@NotNull File dataFolder,
                                                                                   @NotNull String language) {
        File strings = new File(dataFolder, "strings.yml");
        if (!strings.exists()) {
            Bukkit.getLogger().log(Level.FINEST, "Strings file not found");
            return null;
        }

        try {
            Bukkit.getLogger().log(Level.INFO, "Loading custom strings...");
            return loadTranslatableMessages(language,
                    FileHelper.getBufferedReaderFromString(strings.toPath().toString()));
        } catch (FileNotFoundException exception) {
            Bukkit.getLogger().log(Level.WARNING, "Unable to load custom messages");
            return null;
        }
    }

    /**
     * Loads translatable messages from the given reader
     *
     * @param language <p>The selected language</p>
     * @param reader   <p>The buffered reader to read from</p>
     * @return <p>The loaded translated strings</p>
     */
    private @NotNull Map<TranslatableMessage, String> loadTranslatableMessages(@NotNull String language,
                                                                               @NotNull BufferedReader reader) {
        Map<TranslatableMessage, String> translatedMessages = new HashMap<>();
        YamlConfiguration configuration = YamlConfiguration.loadConfiguration(reader);

        for (TranslatableMessage translatableMessageCategories : messageCategories) {
            for (TranslatableMessage translatableMessage : translatableMessageCategories.getAllMessages()) {
                String translated = configuration.getString(language + "." + translatableMessage.name());
                if (translated != null) {
                    translatedMessages.put(translatableMessage, translated);
                }
            }
        }
        return translatedMessages;
    }

}